#if !defined HATP_PLAN_HH_
#define HATP_PLAN_HH_

#include<string>

#include "TaskListNode.hh"
#include "TaskList.hh"
#include "PlanTree.hh"
#include "TimeProjection.hh"
#include "WorldBase.hh"
#include "SocialRuleBase.hh"
#include "PartialPlanAppraiser.hh"

#define COSTS_PART "costs"
#define TIME_PART "time"
#define EFFORT_BALANCING_PART "effort_balancing"
#define INTRICACY_CONTROL_PART "intricacy_control"
#define WASTED_TIME_PART "wasted_time"

class Plan {

private:
	unsigned long int planID;
	TaskList projection;
	std::vector<unsigned long int> TaskProjectionOrder;
	TimeProjection* projectionTimeCanvas;
	PlanTree tree;
	PlanTree abstraction;
	float costs;
	float penalties;
	std::map<std::string, float> scoreDetails;

public:
	static std::map<std::string, float> scoreWeights;
	static unsigned long int planCount;
	explicit Plan(TaskList const& ppj, std::vector<unsigned long int> const& projection_order, PlanTree const& pt, PartialPlanAppraiser const& ppa);
	void printPlan() const;
	bool postTreatments(LightWorldBase& wb_at_start, SocialRuleBase const& srb, float scoreMax);
	PlanTree const& getTree() const;
	PlanTree const& getAbstraction() const;
	TaskList const& getProjection() const;
	TimeProjection* getTimeProjection() const;
	std::vector<unsigned long int> const& getTaskProjectionOrder() const;
	float getScore() const;
	std::map<std::string, float> const& getScoreDetails() const;
	std::map<std::string, float> const* getPtrScoreDetails() const;
	float determineContribution(std::map<std::string, float> const& detailsRef) const;
	unsigned long int getID() const;
	~Plan();
};

/**
 * constructor for rough plan: make a copy of ppj and pt by calling their, argScore must be the projection cost of ppj respective constructors by copying,
 * sets score to argScore. Then, erase useless links in the projection, erase empty branch, propagate involved agents in the tree and make the abstraction.
 */
inline Plan::Plan(TaskList const& ppj, std::vector<unsigned long int> const& projection_order, PlanTree const& pt, PartialPlanAppraiser const& ppa):
	projection(ppj), TaskProjectionOrder(projection_order), projectionTimeCanvas(NULL), tree(pt), abstraction(pt) {
	planID = planCount++;
	// we make some treatments on tree to erase empty branches and to propagate involved agents
	tree.eraseEmptyBranches();
	tree.propagateInvolvedAgent();
	// we make the abstraction from tree
	abstraction.eraseEmptyBranches();
	abstraction.propagateInvolvedAgent();
	abstraction.convertToAbstraction();
	// we make some treatments on the projection and we create projectionTimeCanvas
	projection.destroyAllLinksOfType(HIGH_LEVEL_LINK);
	// we update scoreDetails
	std::map<std::string,float> const& badDec = ppa.getBadDecompoPenalties();
	std::map<std::string,float> const& undSeq = ppa.getUndSeqPenalties();
	std::map<std::string,float>::const_iterator it;
	for(it = badDec.begin() ; it != badDec.end() ; ++it)scoreDetails[it->first] = it->second;
	for(it = undSeq.begin() ; it != undSeq.end() ; ++it) scoreDetails[it->first] = it->second;
	scoreDetails[COSTS_PART] = ppa.getCurrentProjectionCost();
}

/**
 * return tree
 */
inline PlanTree const& Plan::getTree() const {
	return tree;
}

/**
 * return abstraction
 */
inline PlanTree const& Plan::getAbstraction() const {
	return abstraction;
}

/**
 * return projection
 */
inline TaskList const& Plan::getProjection() const {
	return projection;
}

/**
 * return timeProjection
 */
inline TimeProjection* Plan::getTimeProjection() const {
	return projectionTimeCanvas;
}

/**
 * return TaskProjectionOrder
 */
inline std::vector<unsigned long int> const& Plan::getTaskProjectionOrder() const {
	return TaskProjectionOrder;
}

/**
 * compute score from scoreDetails and scoreWeights contents
 */
inline float Plan::getScore() const {
	float score=0.;
	std::map<std::string, float>::const_iterator it, it1;
	for(it = scoreDetails.begin() ; it != scoreDetails.end() ; ++it) {
		it1 = scoreWeights.find(it->first);
		score += it->second * (it1->second);
	}
	return score;
}

/**
 * return scoreDetails
 */
inline std::map<std::string, float> const& Plan::getScoreDetails() const {
	return scoreDetails;
}

/**
 * return &scoreDetails
 */
inline std::map<std::string, float> const* Plan::getPtrScoreDetails() const {
	return &scoreDetails;
}

inline unsigned long int Plan::getID() const {
	return planID;
}

#endif
