#if !defined HATP_TASK_INSTANCE_HH_
#define HATP_TASK_INSTANCE_HH_

#include<vector>
#include<string>
#include<map>

#include <Task.hh>
#include <WorldBaseReport.hh>

class TaskList;
class HTNBase;
class LightWorldBase;
class WorldBase;
class WorldBaseReport;
class hatpPlanningSession;

class TaskInstance {
public:
	enum status{todo, todo_locked, done};
private:
	/**
	* The associated task (i.e. this class is an instance of this general task)
	*/
	Task const& generalTask;
	
	/**
	* Unique ID to differentiate this instance to any other
	* Is the copy of the TaskListNode encapsulating this instance, or -1
	*/
	long uniqueId;

	/**
	* vector of the arguments of the task
	*/
	std::vector<std::string> args;

	/**
	* minimum Time for action realization
	*/
	double minTime;

	/**
	* maximum Time for action realization
	*/
	double maxTime;
	
	/**
	* contains the report generated during precondition checking
	*/
	WorldBaseReport precondReport;
	
	const TaskInstance* originalInstance;
	struct GTPInformation{
		std::pair<int, int> projection;
		double cost;
	};
	GTPInformation gtpInformation;
	
	status currentStatus;
	
public:
	/**
	* Pointer to stock of general tasks
	*/
	static HTNBase* htn_base;

	explicit TaskInstance(std::string const& taskName, std::vector<std::string> const& vecArgs, LightWorldBase const& wb);
	explicit TaskInstance(std::string const& taskName, std::vector<std::string> const& vecArgs, WorldBase const& wb);
	explicit TaskInstance(std::string const& taskName, std::vector<std::string> const& vecArgs, std::map<std::string, std::string> const& varAndTypes);
	TaskInstance(TaskInstance const& ti);
	Task const& getGeneralTask() const;
	void setUniqueId(long uniqueId);
	long getUniqueId() const;
	void setStatus(status newStatus);
	status getStatus() const;
	std::vector<std::string> const& getArgs() const;
	void setHTNBase(HTNBase* argHTNBase);
	bool checkPreconditions(LightWorldBase& wb, const hatpPlanningSession& session);
	void getPreconditionsReport(WorldBaseReport& report) const;
	bool checkTaskGoal(LightWorldBase const& wb, hatpPlanningSession const& session) const;
	std::vector<TaskList*> getDecompositions(LightWorldBase const& wb, HTNBase const& htnb, hatpPlanningSession const& session) const;
	void applyActionEffects(LightWorldBase& wb) const;
	void applyActionEffectsWithReport(LightWorldBase& wb, WorldBaseReport& report) const;
	float ComputeCost(LightWorldBase const& wb) const;
	void ComputeTimeLimit(LightWorldBase const& wb);
	double getMinTime() const;
	double getMaxTime() const;
	std::string toString() const;
	std::vector<std::string> getAssociatedAgents() const;
	bool isAction() const;
	bool operator==(TaskInstance const& ti) const;
	~TaskInstance();
	
	bool isGTPAlternative() const;
	void setGTPAlternative(const TaskInstance* originalInstance);
	const TaskInstance* getGTPOriginal();
	void setGTPInformation(const std::pair<int, int>& projection, double cost);
	const std::pair<int, int>& getGTPProjection() const;
	double getGTPCost() const;
};

#define TI_STATUS TaskInstance::status
#define TODO TaskInstance::todo
#define DONE TaskInstance::done
#define TODO_LOCKED TaskInstance::todo_locked

/**
 * constructor by copying
 */
inline TaskInstance::TaskInstance(TaskInstance const& ti)
	: generalTask(ti.getGeneralTask()), uniqueId(ti.uniqueId),
	args(ti.getArgs()), minTime(ti.getMinTime()),
	maxTime(ti.getMaxTime()), precondReport(ti.precondReport),
	originalInstance(ti.originalInstance), gtpInformation(ti.gtpInformation), currentStatus(ti.currentStatus) {
}

/**
 * return generalTask
 */
inline Task const& TaskInstance::getGeneralTask() const {
	return generalTask;
}

inline void TaskInstance::setUniqueId(long _uniqueId){
	uniqueId=_uniqueId;
}
/**
 * return the unique Id
 */
inline long TaskInstance::getUniqueId() const {
	return uniqueId;
}

inline void TaskInstance::setStatus(status newStatus) {
	currentStatus=newStatus;
}

inline TaskInstance::status TaskInstance::getStatus() const {
	return currentStatus;
}

/**
 * return args
 */
inline std::vector<std::string> const& TaskInstance::getArgs() const {
	return args;
}

/**
 * return generalTask->isAction()
 */
inline bool TaskInstance::isAction() const {
	return generalTask.isAction();
}

/**
 * return minTime
 */
inline double TaskInstance::getMinTime() const {
	return minTime;
}

/**
 * return maxTime
 */
inline double TaskInstance::getMaxTime() const {
	return maxTime;
}

inline bool TaskInstance::isGTPAlternative() const{
	if(!generalTask.isActiveGTP()){
		std::cerr<<"ERROR: Trying to check GTP alternative for the task "<<generalTask.getName()<<" which is not a GTP task"<<std::endl;
		return false;
	}
	return (originalInstance!=NULL);
}

inline void TaskInstance::setGTPAlternative(const TaskInstance* original){
	if(!generalTask.isActiveGTP()){
		std::cerr<<"ERROR: Trying to set GTP original for the task "<<generalTask.getName()<<" which is not a GTP task"<<std::endl;
		return;
	}
	originalInstance=original;
}

inline const TaskInstance* TaskInstance::getGTPOriginal(){
	if(!generalTask.isActiveGTP()){
		std::cerr<<"ERROR: Trying to get GTP original for the task "<<generalTask.getName()<<" which is not a GTP task"<<std::endl;
		return NULL;
	}
	return originalInstance;
}

inline void TaskInstance::setGTPInformation(const std::pair<int, int>& projection, double cost){
	if(!generalTask.isActiveGTP()){
		std::cerr<<"ERROR: Trying to set GTP information for the task "<<generalTask.getName()<<" which is not a GTP task"<<std::endl;
		return;
	}
	gtpInformation.projection=projection;
	gtpInformation.cost=cost;
}

inline const std::pair<int, int>& TaskInstance::getGTPProjection() const{
	if(!generalTask.isActiveGTP()){
		std::cerr<<"ERROR: Trying to get GTP projection for the task "<<generalTask.getName()<<" which is not a GTP task"<<std::endl;
	}
	return gtpInformation.projection;
}

inline double TaskInstance::getGTPCost() const{
	if(!generalTask.isActiveGTP()){
		std::cerr<<"ERROR: Trying to get GTP cost for the task "<<generalTask.getName()<<" which is not a GTP task"<<std::endl;
		return 0.;
	}
	return gtpInformation.cost;
}

/**
 * destructor: empty
 */
inline TaskInstance::~TaskInstance() {}

#endif
