/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>
#include<sstream>
#include<vector>

#include "TaskList.hh"
#include "TaskInstance.hh"
#include "PlanTreeTaskList.hh"
#include "Substitution.hh"
#include "WorldBase.hh"
#include "LightWorldBase.hh"
#include "tools.hh"

using namespace std;

/**
 * write the last projected task (i.e. with the DONE status) associated to Agent ag in lastTsk and return true 
 * if lastTsk has been found, else retrun false (meaning that there no projected task associated to Ag)
 */
bool TaskList::lastProjectedTaskAssociatedToAgent(string const& Ag, unsigned long int& lastTsk) const
{
	using namespace HATP_TOOLS;
	vector<unsigned long int> vec;
	unsigned long int i,j;
	for(i=0 ; i<nodes.size() ; ++i) 
		if(nodes[i]->getStatus() == DONE && Ag >> nodes[i]->getInstance().getAssociatedAgents())
		vec.push_back(nodes[i]->getID());
	if(vec.size() == 0) return false;
	else {
		vector<TaskListNode*> succ;
		for(i=0 ; i<vec.size() ; ++i) {
			succ = this->find(vec[i])->getSuccessors();
			if(succ.size() == 0) { 
				lastTsk = vec[i];
				return true;
			}
			else {
				for(j=0 ; j<succ.size() ; ++j)
					if(succ[j]->getID() >> vec) break;
				if(succ.size()>0 && j == succ.size()) { 
					lastTsk = vec[i];
					return true;
				}
			}
		}
	} 
	return false;
}

/**
 * constructor by copying
 * reviseID controls the ID revision of the nodes
 */
TaskList::TaskList(TaskList const& TL, bool reviseIDs) : nodes2BeDeleted(true)
{
	vector<string> tmp_link_types;
	vector<TaskListNode*> const& vec=TL.getNodes();
	map<TaskListNode*, unsigned int> AdressToRange;
	unsigned int i,j;
	for(i=0 ; i<vec.size() ; ++i) {
		nodes.push_back(new TaskListNode(*vec[i], reviseIDs));
		// we keep correspondance of adress <-> range in nodes vector to make a copy of links
		AdressToRange[vec[i]] = i;
	}
	// we make a copy of all causal links of TL
	for(i=0 ; i<vec.size() ; ++i) {
		// only copy of predecessors (successors will be done AddPredecessors/Successors of TaskListNode)
		vector<TaskListNode*> const& tmp=vec[i]->getPredecessors();
		tmp_link_types=vec[i]->getPredecessorsLinkTypes();
		for(j=0 ; j<tmp.size() ; ++j)
			nodes[i]->AddPredecessor(nodes[AdressToRange[tmp[j]]], tmp_link_types[j]);
	}
}

/**
 * constructor: complete nodes with members of tasks, and update causal links from links member (links
 * contains links for predecessors), all created links are of type HIGH_LEVEL_LINK
 * NB : unsigned integers in the two map does not affect TaskListNode IDs, they are here only to describe causal links (in that sense they must be consistent)
 * NB2 : if IDs != NULL we store in it matching before unsigned int into tasks/links and the IDs of the created TaskListNodes
 * NB3 : map arguments are passed with const reference to save time during execution (it prevents calling constructor by copying)
 */
TaskList::TaskList(map<unsigned int, TaskInstance*> const& tasks,
	map<unsigned int, vector<unsigned int> > const& links, map<unsigned long int, unsigned int>* IDs) : nodes2BeDeleted(true)
{
	map<unsigned int, TaskListNode*> myTLnodes; // to keep the association ID_args <-> TLnodes of current TaskList
	TaskListNode* TLn;
	map<unsigned int, TaskInstance*>::const_iterator it_tasks;
	for(it_tasks=tasks.begin() ; it_tasks != tasks.end() ; ++it_tasks) {
		TLn = new TaskListNode(*(it_tasks->second));
		delete it_tasks->second; // we free memory
		nodes.push_back(TLn);
		myTLnodes[it_tasks->first] = TLn;
		if(IDs != NULL) // we complete IDs if it is necessary (this is useful for problem parsing)
			(*IDs)[TLn->getID()] = it_tasks->first;
	}
	map<unsigned int, vector<unsigned int> >::const_iterator it_links;
	vector<unsigned int> vec;
	unsigned int i;
	for(it_links=links.begin() ; it_links != links.end() ; ++it_links) {
		vec = it_links->second;
		for(i=0 ; i<vec.size() ; ++i)
			myTLnodes[it_links->first]->AddPredecessor(myTLnodes[vec[i]], HIGH_LEVEL_LINK);
	}
	myTLnodes.clear();
	vec.clear();
}

/**
 * constructor: complete nodes with members of tasks, and update causal links from links member (links
 * contains links for predecessors), all created links are of type stored in linkTypes
 * NB : unsigned integers in the two map does not affect TaskListNode IDs, they are here only to describe causal links (in that sense they must be consistent)
 * NB2 : if IDs != NULL we store in it matching between unsigned int into tasks/links/linkTypes and the IDs of the created TaskListNodes
 * NB3 : map arguments are passed with const reference to save time during execution (it prevents calling constructor by copying)
 */
TaskList::TaskList(map<unsigned int, TaskInstance*> const& tasks,
	map<unsigned int, vector<unsigned int> > const& links,
	map<unsigned int, vector<string> > const& linkTypes, map<unsigned long int, unsigned int>* IDs) : nodes2BeDeleted(true)
{
	map<unsigned int, TaskListNode*> myTLnodes; // to keep the association ID_args <-> TLnodes of current TaskList
	TaskListNode* TLn;
	map<unsigned int, TaskInstance*>::const_iterator it_tasks;
	for(it_tasks=tasks.begin() ; it_tasks != tasks.end() ; ++it_tasks) {
		TLn = new TaskListNode(*(it_tasks->second));
		delete it_tasks->second; // we free memory
		nodes.push_back(TLn);
		myTLnodes[it_tasks->first] = TLn;
		if(IDs != NULL)
			(*IDs)[TLn->getID()] = it_tasks->first; // we complete IDs if it is necessary (this is useful for problem parsing)
	}
	map<unsigned int, vector<unsigned int> >::const_iterator it_links;
	vector<unsigned int> vec;
	vector<string> vecTypes;
	unsigned int i;
	for(it_links=links.begin() ; it_links != links.end() ; ++it_links) {
		vec = it_links->second;
		vecTypes = linkTypes.find(it_links->first)->second;
		for(i=0 ; i<vec.size() ; ++i)
			myTLnodes[it_links->first]->AddPredecessor(myTLnodes[vec[i]], vecTypes[i]);
	}
	myTLnodes.clear();
	vec.clear();
	vecTypes.clear();
}

/**
 * constructor: makes a copy of PTTL under a TaskList form, all links are of type HIGH_LEVEL
 * oldAndNewIDs is completed by couples (PTTL ID, TaskList ID) to keep a trace between PlanTree IDs
 * and new TaskListNode IDs.
 */
TaskList::TaskList(PlanTreeTaskList const* PTTL, map<unsigned long int, unsigned long int>& oldAndNewIDs) : nodes2BeDeleted(true)
{
	TaskListNode* TLn;
	vector<PlanTreeTaskListNode*> const& vec=PTTL->getNodes();
	map<PlanTreeTaskListNode*, unsigned int> AdressToRange;
	unsigned int i,j;
	for(i=0 ; i<vec.size() ; ++i) {
		TLn = new TaskListNode(*vec[i]);
		nodes.push_back(TLn);
		oldAndNewIDs[vec[i]->getID()] = TLn->getID();
		AdressToRange[vec[i]] = i; // we keep correspondance of adress <-> range in nodes vector to make a copy of links
	}
	// we make a copy of all causal links of TL
	for(i=0 ; i<vec.size() ; ++i) {
		// we make only a copy of predecessors (copy of successors will be done by method AddPredecessors/Successors of class TaskListNode)
		vector<PlanTreeTaskListNode*> const& tmp=vec[i]->getPredecessors();
		for(j=0 ; j<tmp.size() ; ++j)
			nodes[i]->AddPredecessor(nodes[AdressToRange[tmp[j]]], HIGH_LEVEL_LINK);
	}
}

TaskList* TaskList::makeCopy() const{
	std::vector<TaskListNode*> newNodes;
	std::map<TaskListNode*, TaskListNode*> oldAddressToNewOne;
	
	//First copy the nodes
	for(auto i: nodes){
		newNodes.push_back(i->makeCopy());
		oldAddressToNewOne[i]=newNodes.back();
	}
	
	//Then complete the successors and predecessors
	for(auto i: nodes){
		std::vector<TaskListNode*> const& successors=i->getSuccessors();
		std::vector<std::string> const& linksType=i->getSuccessorsLinkTypes();
		for(size_t j=0; j<successors.size(); ++j){
			oldAddressToNewOne[i]->AddSuccessor(oldAddressToNewOne[successors[j]], linksType[j]);
		}
	}

	return new TaskList(newNodes);
}

/**
 * explore nodes and return the node with the ID equals to argID
 * else return NULL;
 */
TaskListNode* TaskList::find(unsigned long int argID) const
{
	TaskListNode* p = NULL;
	for(unsigned int i=0 ; i<nodes.size() ; ++i) {
		if(nodes[i]->getID()==argID)
		{
			p = nodes[i];
			break;
		}
	}
	return p;
}

/**
 * explore nodes and return the node with the TaskInstance t
 * else return NULL;
 */
TaskListNode* TaskList::find(TaskInstance const& t) const
{
	TaskListNode* p = NULL;
	for(unsigned int i=0 ; i<nodes.size() ; ++i) {
		if(nodes[i]->getInstance()==t) {
			p = nodes[i];
			break;
		}
	}
	return p;
}

/**
 * explore nodes and return all nodes with a TaskInstance using task called
 * taskName
 */
vector<TaskListNode*> TaskList::find(string const& taskName) const
{
	vector<TaskListNode*> vec;
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		if(nodes[i]->getInstance().getGeneralTask().getName() == taskName)
		vec.push_back(nodes[i]);
	return vec;
}

/**
 * explore nodes and return all nodes with a TaskInstance equal to t
 */
vector<TaskListNode*> TaskList::findNodesWithInstance(TaskInstance const& t) const
{
	vector<TaskListNode*> vec;
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		if(nodes[i]->getInstance() == t)
		vec.push_back(nodes[i]);
	return vec;
}

/**
 * return the vector of IDs for all applyable tasks that they have the #TODO or TODO_LOCKED status
 * and that they have no predecessors or predecessors are all done
 */
vector<unsigned long int> TaskList::getFirstTasks(LightWorldBase& wb, const hatpPlanningSession& session)
{
	unsigned int i;
	vector<unsigned long int> vec;
	for(i=0 ; i<nodes.size() ; ++i){
		if((nodes[i]->getStatus() == TODO || nodes[i]->getStatus() == TODO_LOCKED) && (nodes[i]->noPredecessors() || nodes[i]->predecessorsAllDone())){
			if(nodes[i]->getInstance().checkPreconditions(wb, session)) {
				vec.push_back(nodes[i]->getID());
			}
		}
	}
	return vec;
}

/**
 * return the vector of IDs for all applyable tasks that they have the #TODO status
 * and that they have no predecessors are all done predecessors
 */
vector<unsigned long int> TaskList::getFirstUnlockedTasks(LightWorldBase& wb, const hatpPlanningSession& session)
{
	vector<unsigned long int> vec;
	for(unsigned int i=0 ; i<nodes.size() ; ++i){
		if( nodes[i]->getStatus() == TODO && (nodes[i]->noPredecessors() || nodes[i]->predecessorsAllDone())){
			if(nodes[i]->getInstance().checkPreconditions(wb, session)) {
				vec.push_back(nodes[i]->getID());
			}
		}
	}
	return vec;
}

/**
 * substitute TaskListNode with ID = argID by nodes (not a copy, keeps the one passed in TL!) of TaskList* TL,
 * type of created links depends on previous type link
 */
void TaskList::substitute(unsigned long int argID, TaskList* TL)
{
	TaskListNode* TLn = this->find(argID);
	if(TLn == NULL)
		cerr << "ERROR: trying to subsitute a TaskListNode in a TaskList : node has not been found" << endl;
	else {
		if(TLn->getInstance().isAction() && !TLn->getInstance().getGeneralTask().isActiveGTP()) {
			cerr << "ERROR: trying to subsitute a TaskListNode delegating an action ";
			cerr << "by a TaskList (no warranty about the result)" << endl;
		}
		// we explore TLnodes: all nodes with no predecessors takes predecessors of nodes[i]
		// and all nodes with no successors takes successors of nodes[i]
		vector<TaskListNode*> tmpP = TLn->getPredecessors();
		vector<string> tmpP_links = TLn->getPredecessorsLinkTypes();
		vector<TaskListNode*> tmpS = TLn->getSuccessors();
		vector<string> tmpS_links = TLn->getSuccessorsLinkTypes();
		unsigned int j,k;
		if(TL->isEmpty()) { // we have only to remove TLn from the current TaskList
			// we shunt TLn
			for(j=0 ; j<tmpP.size() ; ++j) // all its predecessors becomes successors of its predecessors
				for(k=0 ; k<tmpS.size() ; ++k)
				tmpS[k]->AddPredecessor(tmpP[j], tmpP_links[j]);
		}
		else { // we have to substitute TLn with TL
			vector<TaskListNode*> TLnodes = TL->getNodes();
			for(j=0 ; j<TLnodes.size() ; ++j) {
				if(TLnodes[j]->noPredecessors())
					for(k=0 ; k<tmpP.size() ; ++k) TLnodes[j]->AddPredecessor(tmpP[k], tmpP_links[k]);
				if(TLnodes[j]->noSuccessors())
					for(k=0 ; k<tmpS.size() ; ++k) TLnodes[j]->AddSuccessor(tmpS[k], tmpS_links[k]);
			}
			this->addNodes(TLnodes); // we add TL nodes in nodes
		}
		// we delete TLn from the TaskList (and we destroy of all pointers on it)
		for(k=0 ; k<tmpP.size() ; ++k) // we destroy successors pointers of all TLn predecessors
			tmpP[k]->RemoveSuccessor(TLn);
		for(k=0 ; k<tmpS.size() ; ++k) // we destroy predecessors pointers of all TLn successors
			tmpS[k]->RemovePredecessor(TLn);
		// we update nodes and delete those that are not longer necessary
		vector<TaskListNode*>::iterator it;
		for(it=nodes.begin() ; it!=nodes.end() ; ++it)
			if(*it == TLn) {
				nodes.erase(it); // we remove TLn pointer from nodes
				break;
			}
		delete TLn; // we destroy nodes[i]
	}
}

/**
 * finds task node with ID and sets its status to status
 */
void TaskList::setStatus(unsigned long int ID, TI_STATUS status)
{
	TaskListNode* TLn = this->find(ID);
	if(TLn == NULL)
		cerr << "ERROR: trying to sets the status of a TaskListNode in a TaskList : node has not been found" << endl;
	else
		TLn->setStatus(status);
}

/**
 * set status of all TaskListNode in nodes to status
 */
void TaskList::setAllStatus(TI_STATUS status)
{
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		nodes[i]->setStatus(status);
}

/**
 * converts TaskList in string and returns it
 */
string TaskList::toString() const
{
	ostringstream st;
	if(nodes.size()==0) st << "()" << endl;
	else {
		st << "-----Nodes-----\n";
		unsigned int i,j;
		for(i=0 ; i<nodes.size() ; ++i)
			st << "node " << nodes[i]->getID() << ": " << nodes[i]->toString() << "\n";
		st << "-----Links-----\n";
		vector<TaskListNode*> vec;
		vector<string> vecLinkTypes;
		for(i=0 ; i<nodes.size() ; ++i) {
			vec=nodes[i]->getSuccessors();
			vecLinkTypes=nodes[i]->getSuccessorsLinkTypes();
			for(j=0 ; j<vec.size() ; ++j)
				st << "causal link: from " << nodes[i]->getID() << " to " << vec[j]->getID() << " (type = " << vecLinkTypes[j] << ")" << endl;
		}
		st << "---------------\n";
	}
	string s(st.str());
	return s;
}

/**
 * find tskToProject and project it: i.e. if it is an action and its status is #TODO,
 * set its status to DONE and creates causal links (of type LOW_LEVEL_LINK) due to the affectation of the task
 * to one or more agents (we suppose that an agent can only do one action at a given time).
 */
void TaskList::project(unsigned long int tskToProject, float costTskProjection)
{
	TaskListNode* tskNode = this->find(tskToProject);
	if(tskNode == NULL) { // tskToProject has not been found
		cerr << "ERROR: trying to project task " << tskToProject << " which is not in the TaskList:" << endl;
		cerr << this->toString() << endl << "(operation ignored)" << endl;
	}
	else if(tskNode->getStatus() != TODO) { // tskToProject has not the #TODO status
		cerr << "ERROR: trying to project task " << tskToProject << " which has not the TODO status in the TaskList:" << endl;
		cerr << this->toString() << endl << "(operation ignored)" << endl;
	}
	else { // everything is OK we have to change status of tskToProject and create adapted causal links
		// for each Agent associated to tskToProject, we get all of his projected actions and we make a causal link
		// between the oldest of these actions and tskToProject (only if necessary)
		vector<string> vecTskAgents = tskNode->getInstance().getAssociatedAgents();
		if(vecTskAgents.empty()) {
			cerr << "ERROR: when trying to project the Task " << tskNode->getInstance().toString();
			cerr << ", the system has not found any associated Agent to the task" << endl;
		}
		else {
			unsigned long int lastTask;
			//for each Agent associated to tskToProject, we create causal links between his action stream and tskToProject
			for(unsigned int i=0 ; i<vecTskAgents.size() ; ++i)
				if(this->lastProjectedTaskAssociatedToAgent(vecTskAgents[i], lastTask)) // we get the last projected task of Agent vecTskAgents[i]
				tskNode->AddPredecessor(this->find(lastTask), AGENT_LEVEL_LINK);// we create a link from retainedTask to tskToProject (if it already exists, the function will do nothing)
			// we change status of tskToProject to DONE
			tskNode->setStatus(DONE);
			// we update cost of costTskProjection
			tskNode->setCost(costTskProjection);
		}
	}
}

/**
 * return all efforts done by each agents involved in each done task of nodes
 */
map<string,float> TaskList::getEfforts() const
{
	map<string, float> efforts;
	map<string, float>::iterator it;
	vector<string> ags;
	unsigned int i,j;
	for(i=0 ; i<nodes.size() ; ++i)
		if(nodes[i]->getStatus() == DONE) {
			ags = nodes[i]->getInstance().getAssociatedAgents();
			for(j=0 ; j<ags.size() ; ++j) {
				it = efforts.find(ags[j]);
				if(it != efforts.end())
					efforts[ags[j]] += (nodes[i]->getCost())/(ags.size());
				else
					efforts[ags[j]] = (nodes[i]->getCost())/(ags.size());
			}
		}
	ags.clear();
	return efforts;
}

/**
 * return the number of crossed links, i.e. the links starting from a time-line and
 * finishing in another time-line
 */
multimap<vector<LightEntity const*>, vector<LightEntity const*> > TaskList::getIntricateLinks(LightWorldBase const& wb) const
{
	using namespace HATP_TOOLS;
	unsigned int i,j,k;
	vector<TaskListNode*> succ;
	vector<string> src, dest;
	bool linkDetected;
	multimap<vector<LightEntity const*>, vector<LightEntity const*> > intricateLinks;
	vector<LightEntity const*> startAgents, endAgents;
	for(i=0 ; i<nodes.size() ; ++i) {
		succ = nodes[i]->getSuccessors();
		src = nodes[i]->getInstance().getAssociatedAgents();
		for(j=0 ; j<succ.size() ; ++j) {
			dest = succ[j]->getInstance().getAssociatedAgents();
			// a crossed link exists if the intersection between nodes[i] agents and succ[j] agents is an empty set
			linkDetected = true;
			for(k=0 ; k<dest.size() ; ++k)
				if( dest[k] >> src ) {
					linkDetected = false;
					break;
				}
			if(linkDetected) {
				startAgents.clear();
				endAgents.clear();
				for(k=0 ; k<src.size() ; ++k) startAgents.push_back(wb.find(src[k], AGENT));
				for(k=0 ; k<dest.size() ; ++k) endAgents.push_back(wb.find(src[k], AGENT));
				intricateLinks.insert(make_pair(startAgents, endAgents));
			}
		}
	}
	return intricateLinks;
}

/**
 * return the number of crossed links starting from a set of time-line startSet and finishing
 * to a set of time-line destSet. If startStrict is true the starting set of time-line must
 * be exactly startSet, else the starting set of time-line must include at least startSet. Idem
 * for endSet.
 */
multimap<vector<LightEntity const*>, vector<LightEntity const*> > TaskList::getIntricateLinksFromTo(LightWorldBase const& wb, vector<string>& startSet, bool startStrict, vector<string>& destSet, bool destStrict) const
{
	using namespace HATP_TOOLS;
	unsigned int i,j,k;
	vector<TaskListNode*> succ;
	vector<string> src, dest;
	bool linkDetected;
	multimap<vector<LightEntity const*>, vector<LightEntity const*> > intricateLinks;
	vector<LightEntity const*> startAgents, endAgents;
	for(i=0 ; i<nodes.size() ; ++i) {
		succ = nodes[i]->getSuccessors();
		src = nodes[i]->getInstance().getAssociatedAgents();
		if( (startStrict && src == startSet) || (!startStrict && isSubVector(startSet, src)) ) {
			for(j=0 ; j<succ.size() ; ++j) {
				dest = succ[j]->getInstance().getAssociatedAgents();
				if( (destStrict && src == destSet) || (!destStrict && isSubVector(destSet, dest)) ) {
					// a crossed link exists if the intersection between nodes[i] agents and succ[j] agents is an empty set
					linkDetected = true;
					for(k=0 ; k<dest.size() ; ++k)
						if( dest[k] >> src ) {
							linkDetected = false;
							break;
						}
					if(linkDetected) {
						startAgents.clear();
						endAgents.clear();
						for(k=0 ; k<src.size() ; ++k) startAgents.push_back(wb.find(src[k], AGENT));
						for(k=0 ; k<dest.size() ; ++k) endAgents.push_back(wb.find(src[k], AGENT));
						intricateLinks.insert(make_pair(startAgents, endAgents));
					}
				}
			}
		}
	}
	return intricateLinks;
}

/**
 * return the number of links starting from a set containing at least one member of robots
 * and finishing in a set containing at least one member which is not in robots
 * only links of type LOW_LEVEL_LINK between nodes with status == DONE are counted
 *
 * It also counts the number of joint-actions and apply a penalty of 2^(n-1) (where n
 * is the number of agents invovled in the action)
 */
unsigned int TaskList::getNbIntricateLinks(set<string> const& robots) const
{
	using namespace HATP_TOOLS;
	unsigned int i,j,k, nb=0;
	vector<string> src, dest;
	bool linkDetected;
	for(i=0 ; i<nodes.size() ; ++i) {
		if(nodes[i]->getStatus() == DONE) {
			vector<TaskListNode*> const& succ = nodes[i]->getSuccessors();
			vector<string> const& succLinks = nodes[i]->getSuccessorsLinkTypes();
			src = nodes[i]->getInstance().getAssociatedAgents();
			for(j=0 ; j<succ.size() ; ++j) {
				if(succ[j]->getStatus() == DONE && (succLinks[j] == LOW_LEVEL_LINK || succLinks[j] == AGENT_LEVEL_LINK)) {
					dest = succ[j]->getInstance().getAssociatedAgents();
					// a crossed link exists if the intersection between nodes[i] agents and succ[j] agents is an empty set
					linkDetected = false;
					for(k=0 ; k<src.size() ; ++k)
						if( src[k] >> robots ) {
							linkDetected = true;
							break;
						}
					if(linkDetected) {
						for(k=0 ; k<dest.size() ; ++k)
							if( dest[k] ^= robots ) {
								nb++;
								break;
							}
					}
				} // end if(succ[j]->getStatus() == DONE)
			}
		} // end if(nodes[i]->getStatus() == DONE)
	}
	
	//Add the joint-actions
	for(i=0 ; i<nodes.size() ; ++i) {
		if(nodes[i]->getStatus() == DONE) {
			vector<string> involvedAgents(nodes[i]->getInstance().getAssociatedAgents());
			if(involvedAgents.size()>1){
				//Add 2^(n-1) to nb, which is equivalent to n-1 shift to the left of 1;
				nb+=(1<<(involvedAgents.size()-1));
			}
		} // end if(nodes[i]->getStatus() == DONE)
	}
	return nb;
}

/**
 * delete all links with tag sLinkType in the TaskList
 */
void TaskList::destroyAllLinksOfType(const string& sLinkType)
{
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		nodes[i]->destroyLinksOfType(sLinkType);
}

/**
 * return true if a path exists from IDstart until IDend
 */
bool TaskList::isThereAPathFromTo(unsigned int IDstart, unsigned int IDend) const
{
	bool modif;
	unsigned int i,j;
	map<TaskListNode*, unsigned int> TLNordersInVectors;
	TaskListNode *TLstart=this->find(IDstart), *TLend=this->find(IDend);
	if( TLstart == NULL || TLend == NULL )
		cerr << "ERROR: trying to find a path between two nodes in a TaskList, at least one of these nodes is NULL" << endl;
	else {
		// we create a map associating TaskListNode* and their order in the vectors
		for(i=0 ; i<nodes.size() ; ++i)
			TLNordersInVectors[nodes[i]] = i;
		// we init connected
		vector<bool> connected(nodes.size(), false);
		connected[TLNordersInVectors[TLstart]] = true;
		do {
			modif = false;
			for(i=0 ; i<nodes.size() ; ++i)
				if(connected[TLNordersInVectors[nodes[i]]]) { // if the node is connected all its successors are also connected
					vector<TaskListNode*> const& succ = nodes[i]->getSuccessors();
					for(j=0 ; j<succ.size() ; ++j)
						if( !connected[TLNordersInVectors[succ[j]]] ) {
							connected[TLNordersInVectors[succ[j]]] = true;
							modif = true;
						}
				}
			if(connected[TLNordersInVectors[TLend]])
				return true;
		} while(modif);
	}
	// propagation in connected achieved and TLend is not connected
	return false;
}

/**
 * return true if the subtasks list subTL is detected in the current TaskList
 * i.e. all nodes of subTL can be found in the current TaskList and all link
 * types are respected (subTL must contains only links of type IMMEDIATE and
 * LATER cf. TaskListNode class)
 */
unsigned int TaskList::checkSubList(TaskList const& subTL, map<unsigned long int, pair<double, double> > const& ptimes) const
{
	unsigned int i,j,k;
	vector<TaskListNode*> const& subTLnodes = subTL.getNodes();
	map<TaskListNode*, vector<TaskListNode*> > subTLcandidatesInTL;
	// we get all candidates in TL for subTL nodes
	for(i=0 ; i<subTLnodes.size() ; ++i) {
		subTLcandidatesInTL[subTLnodes[i]] = vector<TaskListNode*>(0);
		// we explore nodes to find a node with same taskInstance as subTLnodes[i]
		for(j=0 ; j<nodes.size() ; ++j)
			if( subTLnodes[i]->getInstance() == nodes[j]->getInstance() )
				subTLcandidatesInTL[subTLnodes[i]].push_back(nodes[j]);
	}
	// we compute all possible combinations by taking two by two
	// the vector members of subTLcandidatesInTL
	vector< map<TaskListNode*, TaskListNode*> > combinations(0), tmp_combinations;
	map<TaskListNode*, TaskListNode*> current_combination;
	map<TaskListNode*, vector<TaskListNode*> >::iterator it;
	for(it = subTLcandidatesInTL.begin() ; it != subTLcandidatesInTL.end() ; ++it) {
		// initialization
		if(it == subTLcandidatesInTL.begin())
			for(i=0 ; i<(it->second).size() ; ++i) {
				combinations.push_back(map<TaskListNode*, TaskListNode*>());
				combinations[i][it->first] = (it->second)[i];
			}
		else {
			// we update compute tmp_combinations from combinations and it->second
			tmp_combinations.clear();
			for(i=0 ; i<(it->second).size() ; ++i)
				for(j=0 ; j<combinations.size() ; ++j) {
					current_combination = combinations[j];
					current_combination[it->first] = (it->second)[i];
					tmp_combinations.push_back(current_combination);
				}
			combinations = tmp_combinations;
		}
	}
	// for all members in combinations we check the links
	unsigned int nb=0;
	bool good_combination;
	for(i=0 ; i<combinations.size() ; ++i) {
		good_combination = true;
		for(j=0 ; j<subTLnodes.size() && good_combination ; ++j) {
			vector<TaskListNode*> const& succ = subTLnodes[j]->getSuccessors();
			vector<string> const& succLinks = subTLnodes[j]->getSuccessorsLinkTypes();
			for(k=0 ; k<succ.size() ; ++k) {
				if( succLinks[k] == IMMEDIATE_LINK ) { // we check if there is a direct link between TaskList nodes
					if( !(combinations[i][subTLnodes[j]]->isAPredecessorOf(combinations[i][succ[k]])) ) {
						good_combination = false;
						break;
					}
					else {
						pair<double, double> w1 = ptimes.find(combinations[i][subTLnodes[j]]->getID())->second;
						pair<double, double> w2 = ptimes.find(combinations[i][succ[k]]->getID())->second;
						if(w1.first < w2.first - combinations[i][succ[k]]->getInstance().getMinTime()) {
							good_combination = false;
							break;
						}
					}
				}
				else { 
					if( succLinks[k] == LATER_LINK ) { // we check if there is a path in the TaskList from subTLnodes[i] to succ[j]
						if( !(this->isThereAPathFromTo(combinations[i][subTLnodes[j]]->getID(), combinations[i][succ[k]]->getID())) ) {
							good_combination = false;
							break;
						}
					}
					else { // error link type is not recognized
						cerr << "ERROR when checking if a TaskList contains a subTaskList, link type \"" << succLinks[k] << "\" unrecognized" << endl;
						return 0;
					}
				}
			}
		}
		if(good_combination) nb++;
	}
	// no combination OK so we return false
	return nb;
}

/**
 * explore nodes and return all possible Subsitutions according subTL nodes
 * arguments
 */
vector<Substitution*> TaskList::getSubstitutionsOnNodes(TaskList const& subTL, vector<string> const& varNames) const
{
	unsigned int i,j,k,n;
	vector<TaskListNode*> const& subTLnodes = subTL.getNodes();
	vector<TaskListNode*> correspondingNodes;
	vector<Substitution*> vecSub, tmpVecSub;
	Substitution* subst;
	bool compatible;
	for(i=0 ; i<subTLnodes.size() ; ++i) {
		correspondingNodes = this->find(subTLnodes[i]->getInstance().getGeneralTask().getName());
		if(correspondingNodes.empty()) { // it misses a node we return an empty vector
			for(k=0 ; k<vecSub.size() ; ++k) // we free memory
				delete vecSub[k];
			return vector<Substitution*>(0);
		}
		else {// we have found at least one corresponding node in current TL
			vector<string> const& vecSubTLNodesArgs = subTLnodes[i]->getInstance().getArgs();
			if(vecSub.empty()) { // we complete vecSub with possible Substitutions
				for(j=0 ; j<correspondingNodes.size() ; ++j) {
					subst = new Substitution(varNames);
					vector<string> const& vecArgs = correspondingNodes[j]->getInstance().getArgs();
					compatible = true;
					// we are sure that vecSubTLNodesArgs and vecArgs have the same size
					// because for a given task there is a fixed number of arguments
					for(k=0 ; k<vecArgs.size() ; ++k) {
						if( !(subst->isInstancied(vecSubTLNodesArgs[k])) ) // we set it into subst
							subst->setValue(vecSubTLNodesArgs[k], vecArgs[k]);
						else if( subst->getValue(vecSubTLNodesArgs[k]) != vecArgs[k] ) { // subst is not compatible
							delete subst;
							compatible = false;
							break;
						}
					}
					// if compatible is true, we add subst in vecSub
					if(compatible) tmpVecSub.push_back(subst);
				}// end for loop to explore correspondingNodes
			} // end vecSub is empty
			else { // start vecSub is not empty
				for(n=0 ; n<vecSub.size() ; ++n) {
					for(j=0 ; j<correspondingNodes.size() ; ++j) {
						subst = new Substitution(*vecSub[n]);
						vector<string> const& vecArgs2 = correspondingNodes[j]->getInstance().getArgs();
						compatible = true;
						// we are sure that vecSubTLNodesArgs and vecArgs2 have the same size
						// because for a given task there is a fixed number of arguments
						for(k=0 ; k<vecArgs2.size() ; ++k) {
							if( !(subst->isInstancied(vecSubTLNodesArgs[k])) ) // we set it into subst
								subst->setValue(vecSubTLNodesArgs[k], vecArgs2[k]);
							else if( subst->getValue(vecSubTLNodesArgs[k]) != vecArgs2[k] ) { // subst is not compatible
								delete subst;
								compatible = false;
								break;
							}
						}
						// if compatible is true, we add subst in vecSub
						if(compatible) tmpVecSub.push_back(subst);
					} // end for loop to explore correspondingNodes
				} // end for loop to explore vecSub
			} // end vecSub is not empty
		} // end there is at least one corresponding node in current TL for subTLnodes[i]
		for(k=0 ; k<vecSub.size() ; ++k) // we free memory
			delete vecSub[k];
		if(tmpVecSub.empty()) // no Substitution available we return and empty vector
			return vector<Substitution*>(0);
		else {
			vecSub = tmpVecSub;
			tmpVecSub.clear();
		}
	} // end for loop to explore subTLnodes
	return vecSub;
}

/**
 * explore nodes and instantiate them with subst to create an instantied version of the
 * current TaskList, create this instance with new operator and return it
 * (NB: it is user responsibility to delete the returned TaskList when it becomes useless)
 */
TaskList* TaskList::Rebuild(Substitution const& subst, LightWorldBase const& wb) const
{
	unsigned int i,j;
	map<TaskListNode*, TaskListNode*> currentNodesToNewNodes;
	vector<string> args;
	vector<TaskListNode*> exitNodes(nodes.size());
	// we produce exit Nodes
	for(i=0 ; i<nodes.size() ; ++i) {
		vector<string> const& nodeArgs = nodes[i]->getInstance().getArgs();
		args.resize(nodeArgs.size());
		for(j=0 ; j<nodeArgs.size() ; ++j) args[j] = subst.getValue(nodeArgs[j]);
		exitNodes[i] = new TaskListNode(TaskInstance(nodes[i]->getInstance().getGeneralTask().getName(), args, wb));
		currentNodesToNewNodes[nodes[i]] = exitNodes[i];
	}
	// we compute links
	vector<string> tmp_link_types;
	for(i=0 ; i<nodes.size() ; ++i) {
		vector<TaskListNode*> const& tmp=nodes[i]->getPredecessors();
		tmp_link_types=nodes[i]->getPredecessorsLinkTypes();
		for(j=0 ; j<tmp.size() ; ++j) exitNodes[i]->AddPredecessor(currentNodesToNewNodes[tmp[j]], tmp_link_types[j]);
	}
	// we return TaskList
	return new TaskList(exitNodes);
}

/**
 * explore nodes and instantie them with subst to create an instantied version of the
 * current TaskList, create this instance with new operator and return it
 * complete nodesToExitNodes to keep correspondence between nodes and nodes of returned TaskList
 * (NB: it is user responsibility to delete the returned TaskList when it becomes useless)
 */
TaskList* TaskList::Rebuild(Substitution const& subst, LightWorldBase const& wb, map<TaskListNode*, TaskListNode*>& nodesToExitNodes) const
{
	unsigned int i,j;
	vector<string> args;
	vector<TaskListNode*> exitNodes(nodes.size());
	// we produce exit Nodes
	for(i=0 ; i<nodes.size() ; ++i) {
		vector<string> const& nodeArgs = nodes[i]->getInstance().getArgs();
		args.resize(nodeArgs.size());
		for(j=0 ; j<nodeArgs.size() ; ++j) args[j] = subst.getValue(nodeArgs[j]);
		exitNodes[i] = new TaskListNode(TaskInstance(nodes[i]->getInstance().getGeneralTask().getName(), args, wb));
		nodesToExitNodes[nodes[i]] = exitNodes[i];
	}
	// we compute links
	vector<string> tmp_link_types;
	for(i=0 ; i<nodes.size() ; ++i) {
		vector<TaskListNode*> const& tmp=nodes[i]->getPredecessors();
		tmp_link_types=nodes[i]->getPredecessorsLinkTypes();
		for(j=0 ; j<tmp.size() ; ++j) exitNodes[i]->AddPredecessor(nodesToExitNodes[tmp[j]], tmp_link_types[j]);
	}
	// we return TaskList
	return new TaskList(exitNodes);
}

/**
 * return the IDs of DONE nodes in the order in which they will be executed
 * if the TaskList is not all done return an empty vector
 */
vector<unsigned long int> TaskList::getOrderedTaskIDs(string const& agentName) const
{
	using namespace HATP_TOOLS;
	unsigned long int current;
	vector<unsigned long int> vec, result;
	vector<TaskListNode*> pred;
	this->lastProjectedTaskAssociatedToAgent(agentName, current);
	vec.push_back(current);
	if(this->find(current) != NULL) pred = this->find(current)->getPredecessors();
	bool cont = true;
	while(!pred.empty() && cont) {
		cont = false;
		for(unsigned int i = 0 ; i < pred.size() ; i++) {
			if(agentName >> pred[i]->getInstance().getAssociatedAgents()) {
				current = pred[i]->getID();
				cont = true;
			}
		}
		if(cont) {
			vec.push_back(current);
			pred = this->find(current)->getPredecessors();
		}
	}
	for(unsigned int i = vec.size() ; i > 0 ; i--) {
		result.push_back(vec[i-1]);
	}
	return result;
}

/**
 * assignment operator, make a copy of t
 */
TaskList& TaskList::operator=(TaskList const& TL)
{
	unsigned int i;
	// we erase nodes
	for(i=0 ; i<nodes.size() ; ++i) delete nodes[i];
	nodes.clear();
	// we complete nodes with t nodes
	vector<string> tmp_link_types;
	vector<TaskListNode*> vec=TL.getNodes();
	map<TaskListNode*, unsigned int> AdressToRange;
	unsigned int j;
	for(i=0 ; i<vec.size() ; ++i) {
		TaskListNode *node = new TaskListNode(*vec[i], false);
		nodes.push_back(node);
		// we keep correspondance of adress <-> range in nodes vector to make a copy of links
		AdressToRange[vec[i]] = i;
	}
	// we make a copy of all causal links of TL
	for(i=0 ; i<vec.size() ; ++i) {
		// we make only a copy of predecessors (copy of successors will be done
		// by method AddPredecessors/Successors of class TaskListNode)
		vector<TaskListNode*> const& tmp=vec[i]->getPredecessors();
		tmp_link_types=vec[i]->getPredecessorsLinkTypes();
		for(j=0 ; j<tmp.size() ; ++j) nodes[i]->AddPredecessor(nodes[AdressToRange[tmp[j]]], tmp_link_types[j]);
	}
	return *this;
}

/**
 * returns true if the TaskList is equals to t (this operator does not
 * consider IDs to compare the TaskLists)
 */
bool TaskList::operator==(TaskList const& t) const
{
	unsigned int i,j;
	vector<TaskListNode*> const& Tnodes = t.getNodes();
	if(nodes.size() == Tnodes.size()) {
		string sNodes_i;
		for(i=0 ; i<nodes.size() ; ++i) {
			sNodes_i = nodes[i]->getInstance().toString();
			for(j=0 ; j<Tnodes.size() ; ++j) // we have to find a node with same task, predecessors and successors
				if( sNodes_i == Tnodes[j]->getInstance().toString() && nodes[i]->hasSamePredecessors(Tnodes[j]) &&
					nodes[i]->hasSameSuccessors(Tnodes[j]) ) break;
			if(j == Tnodes.size())
				return false;
		}
		return true;
	}
	return false;
}

/**
 * destructor: free memory by deleting all members of nodes if nodes2BeDeleted is true
 * and clear nodes
 */
TaskList::~TaskList()
{
	if(nodes2BeDeleted) {
//		for(unsigned int i=0 ; i<nodes.size() ; ++i)
//			delete nodes[i]; // if nodes[i] is NULL, no effect
	}
}
