/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_SOCIAL_RULE_BASE_HH_
#define HATP_SOCIAL_RULE_BASE_HH_

#include<string>
#include<vector>
#include<list>
#include<map>

#include "TaskList.hh"

class LightWorldBase;
class TimeProjection;
class PlanTree;

typedef std::pair<std::vector<TaskList*>, std::pair<std::map<TaskListNode*, std::pair<TaskList*, std::string> >, std::vector<std::string> > > badDecompoDescription;

typedef float (*undesirableStateFunc)(LightWorldBase const&, TimeProjection const&);
typedef std::pair<TaskList*, std::vector<std::string> > (*undesirableSequenceInitFunc)();
typedef float (*undesirableSequenceFunc)(TaskList const&, std::vector<std::string> const&, LightWorldBase const&, TaskList const&, TimeProjection const&);
typedef float (*effortBalancingFunc)(LightWorldBase const&, std::map<std::string, float> const& efforts);
typedef std::set<std::string> (*intricacyControlInitFunc)();
typedef float (*intricacyControlFunc)(TaskList const&, std::set<std::string> const&);
typedef std::set<std::string> (*wastedTimeInitFunc)();
typedef float (*wastedTimeFunc)(TaskList const&, TimeProjection const&, std::set<std::string> const&);
typedef float (*abstractionLegibilityFunc)(PlanTree const&, PlanTree const&);
typedef badDecompoDescription (*badDecompositionInitFunc)();
typedef float (*badDecompositionFunc)(std::vector<TaskList*> const&, std::map<TaskListNode*, std::pair<TaskList*, std::string> > const&, std::vector<std::string> const&, LightWorldBase const&, PlanTree const&);

/** This class aims at storing the different social ruels (or filtering rules).
 * Those functions are used in the postprocess phase, once a plan is computed, the functions are used to
 * increase or decrease the overall cost of the plan in order to take in account some social rules.
 * This process allows to change the plan considered to be the best.
 */
class SocialRuleBase
{
private:
	/** The list of undesirable state functions.
	*/
	std::map<std::string,undesirableStateFunc> undesirableStates;

	/** The list of undesirable sequence functions.
	*/
	std::map<std::string, std::pair<undesirableSequenceInitFunc, undesirableSequenceFunc> > undesirableSequencesFunctions;

	/** The resulting data of undesirable sequence init functions.
	*/
	std::map<std::string, std::pair<TaskList*, std::vector<std::string> > > undesirableSequencesData;

	/** The function defining the effort balancing rule.
	*/
	effortBalancingFunc effortBalancingRule;

	/** The functions defining the intricacy control rule.
	*/
	std::pair<intricacyControlInitFunc, intricacyControlFunc> intricacyControlFunctions;

	/** The resulting data of intricacy control init function.
	*/
	std::set<std::string> intricacyControlData;

	/** The functions defining the wasted time rule.
	*/
	std::pair<wastedTimeInitFunc, wastedTimeFunc> wastedTimeFunctions;

	/** The resulting data of wasted time init function.
	*/
	std::set<std::string> wastedTimeData;

	/** The function defining the abstraction legibility.
	*/
	abstractionLegibilityFunc abstractionLegibilityRule;

	/** The list of bad decomposition functions.
	*/
	std::map<std::string, std::pair<badDecompositionInitFunc, badDecompositionFunc> > badDecompositionsFunctions;

	/** The resulting data of bad decompo init functions.
	*/
	std::map<std::string, badDecompoDescription> badDecompositionsData;

public:
	explicit SocialRuleBase();
	explicit SocialRuleBase(SocialRuleBase const& srb);
	SocialRuleBase& operator=(SocialRuleBase const& other);
	void init();

	void add(std::string const& name, undesirableStateFunc usf);
	void add(std::string const& name, undesirableSequenceInitFunc usfInit, undesirableSequenceFunc usf);
	void add(effortBalancingFunc ebf);
	void add(intricacyControlInitFunc icfInit, intricacyControlFunc icf);
	void add(wastedTimeInitFunc wtfInit, wastedTimeFunc wtf);
	void add(abstractionLegibilityFunc alf);
	void add(std::string const& name, badDecompositionInitFunc bdfInit, badDecompositionFunc bdf);

	bool effortBalancingRuleDefined() const;
	bool intricacyControlRuleDefined() const;
	bool wastedTimeRuleDefined() const;
	bool abstractionLegibilityRuleDefined() const;

	std::map<std::string,undesirableStateFunc> const& getUndesirableStates() const;
	std::map<std::string, std::pair<undesirableSequenceInitFunc, undesirableSequenceFunc> > const& getUndesirableSequencesFunctions() const;
	std::map<std::string, std::pair<TaskList*, std::vector<std::string> > > const& getUndesirableSequencesData() const;
	effortBalancingFunc getEffortBalancingRule() const;
	std::pair<intricacyControlInitFunc, intricacyControlFunc> getIntricacyControlFunctions() const;
	std::pair<wastedTimeInitFunc, wastedTimeFunc> getWastedTimeFunctions() const;
	abstractionLegibilityFunc getAbstractionLegibilityRule() const;
	std::map<std::string, std::pair<badDecompositionInitFunc, badDecompositionFunc> > const& getBadDecompositionsFunctions() const;
	std::map<std::string, badDecompoDescription> getBadDecompositionsData() const;
	badDecompositionFunc getBadDecompositionFunction(std::string const& name) const;
	undesirableSequenceFunc getUndesirableSequenceFunction(std::string const& name) const;
	std::pair<TaskList*, std::vector<std::string> > getUndesirableSequenceData(std::string const& name) const;
	badDecompoDescription getBadDecompositionData(std::string const& name) const;
	std::set<std::string> const& getIntricacyControlData() const;
	std::set<std::string> const& getWastedTimeData() const;

	~SocialRuleBase();
};

/** Constructor.
 */
inline SocialRuleBase::SocialRuleBase() : effortBalancingRule(NULL), abstractionLegibilityRule(NULL)
{
	intricacyControlFunctions = std::make_pair((intricacyControlInitFunc)NULL, (intricacyControlFunc)NULL);
	wastedTimeFunctions = std::make_pair((wastedTimeInitFunc)NULL, (wastedTimeFunc)NULL);
}

/** Return undesirableStates.
 * @return undesirableStates
 */
inline std::map<std::string,undesirableStateFunc> const& SocialRuleBase::getUndesirableStates() const
{
	return undesirableStates;
}

/** Return undesirableSequencesFunctions.
 * @return undesirableSequencesFunctions
 */
inline std::map<std::string, std::pair<undesirableSequenceInitFunc, undesirableSequenceFunc> > const& SocialRuleBase::getUndesirableSequencesFunctions() const
{
	return undesirableSequencesFunctions;
}

/** Return undesirableSequencesData.
 * @return undesirableSequencesData
 */
inline std::map<std::string, std::pair<TaskList*, std::vector<std::string> > > const& SocialRuleBase::getUndesirableSequencesData() const
{
	return undesirableSequencesData;
}

/** Return effortBalancingRule.
 * @return effortBalancingRule
 */
inline effortBalancingFunc SocialRuleBase::getEffortBalancingRule() const
{
	return effortBalancingRule;
}

/** Return intricacyControlFunctions.
 * @return intricacyControlFunctions
 */
inline std::pair<intricacyControlInitFunc, intricacyControlFunc> SocialRuleBase::getIntricacyControlFunctions() const
{
	return intricacyControlFunctions;
}

/** Return wastedTimeFunctions.
 * @return wastedTimeFunctions
 */
inline std::pair<wastedTimeInitFunc, wastedTimeFunc> SocialRuleBase::getWastedTimeFunctions() const
{
	return wastedTimeFunctions;
}

/** Return intricacyControlRule.
 * @return intricacyControlRule
 */
inline abstractionLegibilityFunc SocialRuleBase::getAbstractionLegibilityRule() const
{
	return abstractionLegibilityRule;
}

/** Return badDecompositionsFunctions.
 * @return badDecompositionsFunctions
 */
inline std::map<std::string, std::pair<badDecompositionInitFunc, badDecompositionFunc> > const& SocialRuleBase::getBadDecompositionsFunctions() const
{
	return badDecompositionsFunctions;
}

/** Return badDecompositionsData.
 * @return badDecompositionsData
 */
inline std::map<std::string, badDecompoDescription> SocialRuleBase::getBadDecompositionsData() const
{
	return badDecompositionsData;
}

/** Set effortBalancingRule.
 * @param ebf The effort balancing function
 */
inline void SocialRuleBase::add(effortBalancingFunc ebf)
{
	effortBalancingRule = ebf;
}

/** Set intricacyControlFunctions.
 * @param icfInit Function to initialise intricacy control
 * @param icf Intricacy function
 */
inline void SocialRuleBase::add(intricacyControlInitFunc icfInit, intricacyControlFunc icf)
{
	intricacyControlFunctions = std::make_pair(icfInit, icf);
}

/** Set intricacyControlFunctions.
 * @param wtfInit Function to initialise the wasted-time function
 * @param wtf Wasted-time function
 */
inline void SocialRuleBase::add(wastedTimeInitFunc wtfInit, wastedTimeFunc wtf)
{
	wastedTimeFunctions = std::make_pair(wtfInit, wtf);
}

/** Set abstractionLegibilityRule.
 * @param Abstraction legibility rule
 */
inline void SocialRuleBase::add(abstractionLegibilityFunc alf)
{
	abstractionLegibilityRule = alf;
}

/** Check if effort balancing is set.
 * @return effortBalancingRule!=NULL
 */
inline bool SocialRuleBase::effortBalancingRuleDefined() const
{
	return effortBalancingRule != NULL;
}

/** Check if intricacy control is set.
 * @return intricacyControlFunctions.first != NULL && intricacyControlFunctions.second != NULL
 */
inline bool SocialRuleBase::intricacyControlRuleDefined() const
{
	return intricacyControlFunctions.first != NULL && intricacyControlFunctions.second != NULL;
}

/** Check if wasted time function is set.
 * @return wastedTimeFunctions.first != NULL && wastedTimeFunctions.second != NULL
 */
inline bool SocialRuleBase::wastedTimeRuleDefined() const
{
	return wastedTimeFunctions.first != NULL && wastedTimeFunctions.second != NULL;
}

/** Check if abstraction legibility rule is set.
 * return abstractionLegibilityRule!=NULL
 */
inline bool SocialRuleBase::abstractionLegibilityRuleDefined() const
{
	return abstractionLegibilityRule != NULL;
}

/** Return a given bad decompostions function.
 * @param name The name of the decomposition
 * @return The badDecomposition if existing, NULL otherwise
 */
inline badDecompositionFunc SocialRuleBase::getBadDecompositionFunction(std::string const& name) const
{
	std::map<std::string, std::pair<badDecompositionInitFunc, badDecompositionFunc> >::const_iterator it = badDecompositionsFunctions.find(name);
	if(it != badDecompositionsFunctions.end())
		return (it->second).second;

	return NULL;
}

/** Return a given undesirable sequence function.
 * @param name The name of the sequence
 * @retrun The undesiravleSequenceFunc if existing, NULL otherwise
 */
inline undesirableSequenceFunc SocialRuleBase::getUndesirableSequenceFunction(std::string const& name) const
{
	std::map<std::string, std::pair<undesirableSequenceInitFunc, undesirableSequenceFunc> >::const_iterator it = undesirableSequencesFunctions.find(name);
	if(it != undesirableSequencesFunctions.end())
		return (it->second).second;

	return NULL;
}

/** Return a given undesirable sequence data.
 * @param name The name of the sequence
 * @retrun The undesirable sequence if existing, empty pair otherwise
 */
inline std::pair<TaskList*, std::vector<std::string> > SocialRuleBase::getUndesirableSequenceData(std::string const& name) const
{
	std::map<std::string, std::pair<TaskList*, std::vector<std::string> > >::const_iterator it = undesirableSequencesData.find(name);
	if(it != undesirableSequencesData.end())
		return it->second;

	return std::make_pair((TaskList*)NULL, std::vector<std::string>(0));
}

/** Return a given bad decomposition data.
 * @param name The name of the decomposition
 * @retrun The bad decompositiom if existing, empty decomposition otherwise
 */
inline badDecompoDescription SocialRuleBase::getBadDecompositionData(std::string const& name) const
{
	std::map<std::string, badDecompoDescription>::const_iterator it = badDecompositionsData.find(name);
	if(it != badDecompositionsData.end())
		return it->second;

	return std::make_pair(std::vector<TaskList*>(0), std::make_pair(std::map<TaskListNode*, std::pair<TaskList*, std::string> >(), std::vector<std::string>(0)));
}

/** Retrun the intricacy control data.
 * @return intricacyControlData
 */
inline std::set<std::string> const& SocialRuleBase::getIntricacyControlData() const
{
	return intricacyControlData;
}

/** Return the wasted time data.
 * @return wastedTimeData
 */
inline std::set<std::string> const& SocialRuleBase::getWastedTimeData() const
{
	return wastedTimeData;
}

/** Destructor.
 * Deletes members of undesirableSequencesData and badDecompositionsData
 */
inline SocialRuleBase::~SocialRuleBase()
{
	std::map<std::string, std::pair<TaskList*, std::vector<std::string> > >::iterator it;
	for(it = undesirableSequencesData.begin() ; it != undesirableSequencesData.end() ; ++it)
		delete (it->second).first;

	std::map<std::string, badDecompoDescription>::iterator it2;
	unsigned int i;
	std::vector<TaskList*> vecTL;
	for(it2 = badDecompositionsData.begin() ; it2 != badDecompositionsData.end() ; ++it2)
	{
		vecTL = (it2->second).first;
		for(i=0 ; i<vecTL.size() ; ++i)
			delete vecTL[i];
	}
}

#endif
